#include "kmsnddev.h"
#include "s_hesad.h"

typedef struct {
	KMIF_SOUND_DEVICE kmif;
	struct {
		Int32 mastervolume;
	} common;
	Uint8 pcmbuf[0x10000];
	Uint8 port[15];
	Uint32 outfreq;
	Uint32 freq;
	Uint16 addr;
	Uint16 writeptr;
	Uint16 readptr;
	Int8 playflag;
	Int8 repeatflag;
	Int32 length;
} HESADPCM;

static void HESAdPcmReset(HESADPCM *sndp)
{
	sndp->addr = 0;
	sndp->freq = 0;
	sndp->writeptr = 0;
	sndp->readptr = 0;
	sndp->playflag = 0;
	sndp->repeatflag = 0;
	sndp->length = 0;
}

static void sndsynth(void *ctx, Int32 *p)
{
	HESADPCM *sndp = ctx;
}

static void sndreset(void *ctx, Uint32 clock, Uint32 freq)
{
	HESADPCM *sndp = ctx;
	XMEMSET(&sndp->pcmbuf, 0, sizeof(sndp->pcmbuf));
	XMEMSET(&sndp->port, 0, sizeof(sndp->port));
	HESAdPcmReset(sndp);
	sndp->outfreq = freq;
}

static void sndwrite(void *ctx, Uint32 a, Uint32 v)
{
	HESADPCM *sndp = ctx;
	switch(a & 15)
	{
	  case 0x8:
		// port low
		sndp->addr &= 0xff00;
		sndp->addr |= v;
		break;
	  case 0x9:
		// port high
		sndp->addr &= 0xff;
		sndp->addr |= v << 8;
		break;
	  case 0xA:
		// write buffer
		sndp->pcmbuf[sndp->writeptr++] = v;
		break;
	  case 0xB:
		// DMA busy?
		break;
	  case 0xC:
		return;
	  case 0xD:
		if (v & 0x80)
		{
			// reset
			HESAdPcmReset(sndp);
		}
		if ((v & 0x03) == 0x03)
		{
			// set write pointer
			sndp->writeptr = sndp->addr;
		}
		if (v & 0x08)
		{
			// set read pointer
			sndp->readptr = sndp->addr;
		}
		if (v & 0x10)
		{
			sndp->length = sndp->addr;
		}
		sndp->repeatflag = ((v & 0x20) == 0x20);
		sndp->playflag = ((v & 0x60) == 0x60);
		break;
	  case 0xE:
		// set freq
		sndp->freq = 32 / (16 - (v & 15));
		break;
	  case 0xF:
		// fade out
		break;
	}
	sndp->port[a & 15] = v;
}

static Uint32 sndread(void *ctx, Uint32 a)
{
	HESADPCM *sndp = ctx;
	switch(a & 15)
	  {
		case 0xa:
			return sndp->pcmbuf[sndp->readptr++];
		case 0xb:
			return sndp->port[0xb] & ~1;
		case 0xc:
			if (!sndp->playflag)
			{
				sndp->port[0xc] |= 1;
				sndp->port[0xc] &= ~8;
			} else {
				sndp->port[0xc] &= ~1;
				sndp->port[0xc] |= 8;
			}
		  return sndp->port[0xc];
		case 0xd:
		  return 0;
		default:
		  return 0xff;
	}
}

#define LOG_BITS 12
static void sndvolume(void *ctx, Int32 volume)
{
	HESADPCM *sndp = ctx;
	volume = (volume << (LOG_BITS - 8)) << 1;
	sndp->common.mastervolume = volume;
}

static void sndrelease(void *ctx)
{
	HESADPCM *sndp = ctx;
	if (sndp)
		XFREE(sndp);
}

static void setinst(void *ctx, Uint32 n, void *p, Uint32 l){}

KMIF_SOUND_DEVICE *HESAdPcmAlloc(void)
{
	HESADPCM *sndp;
	sndp = XMALLOC(sizeof(HESADPCM));
	if (!sndp) return 0;
	XMEMSET(sndp, 0, sizeof(HESADPCM));
	sndp->kmif.ctx = sndp;
	sndp->kmif.release = sndrelease;
	sndp->kmif.reset = sndreset;
	sndp->kmif.synth = sndsynth;
	sndp->kmif.volume = sndvolume;
	sndp->kmif.write = sndwrite;
	sndp->kmif.read = sndread;
	sndp->kmif.setinst = setinst;
	return &sndp->kmif;
}
