/* NSF mapper(4KBx8) */

#include "neserr.h"
#include "handler.h"
#include "audiosys.h"
#include "songinfo.h"
#include "nsf6502.h"
#include "device/nes/s_apu.h"
#include "device/nes/s_mmc5.h"
#include "device/nes/s_vrc6.h"
#include "device/nes/s_vrc7.h"
#include "device/nes/s_fds.h"
#include "device/nes/s_n106.h"
#include "device/nes/s_fme7.h"

#define NSF_MAPPER_STATIC 0
#include "m_nsf.h"

#define EXTSOUND_VRC6	(1 << 0)
#define EXTSOUND_VRC7	(1 << 1)
#define EXTSOUND_FDS	(1 << 2)
#define EXTSOUND_MMC5	(1 << 3)
#define EXTSOUND_N106	(1 << 4)
#define EXTSOUND_FME7	(1 << 5)
#define EXTSOUND_J86	(1 << 6)	/* JALECO-86 */

/* RAM area */
static Uint32 __fastcall ReadRam(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->ram[A & 0x07FF];
}
static void __fastcall WriteRam(void *pNezPlay, Uint32 A, Uint32 V)
{
	((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->ram[A & 0x07FF] = (Uint8)V;
}

/* SRAM area */
static Uint32 __fastcall ReadStaticArea(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->static_area[A - 0x6000];
}
static void __fastcall WriteStaticArea(void *pNezPlay, Uint32 A, Uint32 V)
{
	((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->static_area[A - 0x6000] = (Uint8)V;
}

#if !NSF_MAPPER_STATIC
static Uint32 __fastcall ReadRom8000(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->bank[0][A];
}
static Uint32 __fastcall ReadRom9000(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->bank[1][A];
}
static Uint32 __fastcall ReadRomA000(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->bank[2][A];
}
static Uint32 __fastcall ReadRomB000(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->bank[3][A];
}
static Uint32 __fastcall ReadRomC000(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->bank[4][A];
}
static Uint32 __fastcall ReadRomD000(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->bank[5][A];
}
static Uint32 __fastcall ReadRomE000(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->bank[6][A];
}
static Uint32 __fastcall ReadRomF000(void *pNezPlay, Uint32 A)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->bank[7][A];
}
#endif

/* Mapper I/O */
static void __fastcall WriteMapper(void *pNezPlay, Uint32 A, Uint32 V)
{
	Uint32 bank;
	NSFNSF *nsf = ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf);
	bank = A - 0x5FF0;
	if (bank < 6 || bank > 15) return;
#if !NSF_MAPPER_STATIC
	if (bank >= 8)
	{
		if (V < nsf->banknum)
		{
			nsf->bank[bank - 8] = &nsf->bankbase[V << 12] - (bank << 12);
		}
		else
			nsf->bank[bank - 8] = nsf->zero_area - (bank << 12);
		return;
	}
#endif
	if (V < nsf->banknum)
	{
		XMEMCPY(&nsf->static_area[(bank - 6) << 12], &nsf->bankbase[V << 12], 0x1000);
	}
	else
		XMEMSET(&nsf->static_area[(bank - 6) << 12], 0x00, 0x1000);
}

static NES_READ_HANDLER nsf_mapper_read_handler[] = {
	{ 0x0000,0x0FFF,ReadRam, },
	{ 0x1000,0x1FFF,ReadRam, },
	{ 0x6000,0x6FFF,ReadStaticArea, },
	{ 0x7000,0x7FFF,ReadStaticArea, },
#if !NSF_MAPPER_STATIC
	{ 0x8000,0x8FFF,ReadRom8000, },
	{ 0x9000,0x9FFF,ReadRom9000, },
	{ 0xA000,0xAFFF,ReadRomA000, },
	{ 0xB000,0xBFFF,ReadRomB000, },
	{ 0xC000,0xCFFF,ReadRomC000, },
	{ 0xD000,0xDFFF,ReadRomD000, },
	{ 0xE000,0xEFFF,ReadRomE000, },
	{ 0xF000,0xFFFF,ReadRomF000, },
#else
	{ 0x8000,0x8FFF,ReadStaticArea, },
	{ 0x9000,0x9FFF,ReadStaticArea, },
	{ 0xA000,0xAFFF,ReadStaticArea, },
	{ 0xB000,0xBFFF,ReadStaticArea, },
	{ 0xC000,0xCFFF,ReadStaticArea, },
	{ 0xD000,0xDFFF,ReadStaticArea, },
	{ 0xE000,0xEFFF,ReadStaticArea, },
	{ 0xF000,0xFFFF,ReadStaticArea, },
#endif
	{ 0     ,0     ,0, },
};
static NES_WRITE_HANDLER nsf_mapper_write_handler[] = {
	{ 0x0000,0x0FFF,WriteRam, },
	{ 0x1000,0x1FFF,WriteRam, },
	{ 0x6000,0x6FFF,WriteStaticArea, },
	{ 0x7000,0x7FFF,WriteStaticArea, },
	{ 0     ,0     ,0, },
};
static NES_WRITE_HANDLER nsf_mapper_write_handler2[] = {
	{ 0x5FF6,0x5FFF,WriteMapper, },
	{ 0     ,0     ,0, },
};

Uint8 *NSFGetHeader(NEZ_PLAY *pNezPlay)
{
	return ((NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf)->head;
}

static Uint GetWordLE(Uint8 *p)
{
	return p[0] | (p[1] << 8);
}

static void __fastcall ResetBank(void *pNezPlay)
{
	Uint i, startbank;
	NSFNSF *nsf = (NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf;
	XMEMSET(nsf->ram, 0, 0x0800);
	startbank = GetWordLE(nsf->head + 8) >> 12;
	for (i = 0; i < 16; i++)
		WriteMapper(pNezPlay, 0x5FF0 + i, 0x10000/* off */);
	if (nsf->banksw)
	{
		for (i = 0; (startbank + i) < 8; i++)
			WriteMapper(pNezPlay, 0x5FF0 + (startbank + i) , i);
		if (nsf->banksw & 2)
		{
			WriteMapper(pNezPlay, 0x5FF6, nsf->head[0x70 + 6]);
			WriteMapper(pNezPlay, 0x5FF7, nsf->head[0x70 + 7]);
		}
		for (i = 8; i < 16; i++)
			WriteMapper(pNezPlay, 0x5FF0 + i, nsf->head[0x70 + i - 8]);
	}
	else
	{
		for (i = startbank; i < 16; i++)
			WriteMapper(pNezPlay, 0x5FF0 + i , i - startbank);
	}
}

const static NES_RESET_HANDLER nsf_mapper_reset_handler[] = {
	{ NES_RESET_SYS_FIRST, ResetBank, },
	{ 0,                   0, },
};

static void __fastcall Terminate(void *pNezPlay)
{
	NSFNSF *nsf = (NSFNSF*)((NEZ_PLAY*)pNezPlay)->nsf;
	if (nsf)
	{
		if (nsf->bankbase)
		{
			XFREE(nsf->bankbase);
			nsf->bankbase = 0;
		}
		XFREE(nsf);
		((NEZ_PLAY*)pNezPlay)->nsf = 0;
	}
}

const static NES_TERMINATE_HANDLER nsf_mapper_terminate_handler[] = {
	{ Terminate, },
	{ 0, },
};

static Uint NSFMapperInitialize(NEZ_PLAY *pNezPlay, Uint8 *pData, Uint uSize)
{
	Uint32 size = uSize;
	Uint8 *data = pData;
	NSFNSF *nsf = (NSFNSF*)pNezPlay->nsf;

	size += GetWordLE(nsf->head + 8) & 0x0FFF;
	size  = (size + 0x0FFF) & ~0x0FFF;

	nsf->bankbase = XMALLOC(size + 8);
	if (!nsf->bankbase) return NESERR_SHORTOFMEMORY;

	nsf->banknum = size >> 12;
	XMEMSET(nsf->bankbase, 0, size);
	XMEMCPY(nsf->bankbase + (GetWordLE(nsf->head + 8) & 0x0FFF), data, uSize);

	nsf->banksw = nsf->head[0x70] || nsf->head[0x71] || nsf->head[0x72] || nsf->head[0x73] || nsf->head[0x74] || nsf->head[0x75] || nsf->head[0x76] || nsf->head[0x77];
	if (SONGINFO_GetExtendDevice(pNezPlay->song) & EXTSOUND_FDS) nsf->banksw <<= 1;
	return NESERR_NOERROR;
}

Uint NSFDeviceInitialize(NEZ_PLAY *pNezPlay)
{
	NSFNSF *nsf = (NSFNSF*)pNezPlay->nsf;
	NESResetHandlerInstall(pNezPlay->nrh, nsf_mapper_reset_handler);
	NESTerminateHandlerInstall(&pNezPlay->nth, nsf_mapper_terminate_handler);
	NESReadHandlerInstall(pNezPlay, nsf_mapper_read_handler);
	NESWriteHandlerInstall(pNezPlay, nsf_mapper_write_handler);
	if (nsf->banksw) NESWriteHandlerInstall(pNezPlay, nsf_mapper_write_handler2);

	XMEMSET(nsf->ram, 0, 0x0800);
	XMEMSET(nsf->static_area, 0, sizeof(nsf->static_area));
#if !NSF_MAPPER_STATIC
	XMEMSET(nsf->zero_area, 0, sizeof(nsf->zero_area));
#endif

	APUSoundInstall(pNezPlay);
	if (SONGINFO_GetExtendDevice(pNezPlay->song) & EXTSOUND_VRC6) VRC6SoundInstall(pNezPlay);
	if (SONGINFO_GetExtendDevice(pNezPlay->song) & EXTSOUND_VRC7) VRC7SoundInstall(pNezPlay);
	if (SONGINFO_GetExtendDevice(pNezPlay->song) & EXTSOUND_FDS)  FDSSoundInstall(pNezPlay);
	if (SONGINFO_GetExtendDevice(pNezPlay->song) & EXTSOUND_MMC5)
	{
		MMC5SoundInstall(pNezPlay);
		MMC5MutiplierInstall(pNezPlay);
		MMC5ExtendRamInstall(pNezPlay);
	}
	if (SONGINFO_GetExtendDevice(pNezPlay->song) & EXTSOUND_N106) N106SoundInstall(pNezPlay);
	if (SONGINFO_GetExtendDevice(pNezPlay->song) & EXTSOUND_FME7) FME7SoundInstall(pNezPlay);
	return NESERR_NOERROR;
}

static void NSFMapperSetInfo(NEZ_PLAY *pNezPlay, Uint8 *pData)
{
	XMEMCPY(((NSFNSF*)pNezPlay->nsf)->head, pData, 0x80);
	SONGINFO_SetStartSongNo(pNezPlay->song, pData[0x07]);
	SONGINFO_SetMaxSongNo(pNezPlay->song, pData[0x06]);
	SONGINFO_SetExtendDevice(pNezPlay->song, pData[0x7B]);
	SONGINFO_SetInitAddress(pNezPlay->song, GetWordLE(pData + 0x0A));
	SONGINFO_SetPlayAddress(pNezPlay->song, GetWordLE(pData + 0x0C));
	SONGINFO_SetChannel(pNezPlay->song, 1);
}

Uint NSFLoad(NEZ_PLAY *pNezPlay, Uint8 *pData, Uint uSize)
{
	Uint ret;
	NSFNSF *THIS_ = (NSFNSF *)XMALLOC(sizeof(NSFNSF));
	if (!THIS_) return NESERR_SHORTOFMEMORY;
	XMEMSET(THIS_, 0, sizeof(NSFNSF));
	THIS_->fds_type = 2;
	pNezPlay->nsf = THIS_;
	NESMemoryHandlerInitialize(pNezPlay);
	NSFMapperSetInfo(pNezPlay, pData);
	ret = NSF6502Install(pNezPlay);
	if (ret) return ret;
	ret = NSFMapperInitialize(pNezPlay, pData + 0x80, uSize - 0x80);
	if (ret) return ret;
	ret = NSFDeviceInitialize(pNezPlay);
	if (ret) return ret;
	SONGINFO_SetSongNo(pNezPlay->song, SONGINFO_GetStartSongNo(pNezPlay->song));
	return NESERR_NOERROR;
}
